from collections.abc import Callable
from pathlib import PurePosixPath
from typing import Any, cast

from mkdocs.utils.meta import get_data
from natsort import natsort_keygen, ns
from wcmatch import glob

from mkdocs_awesome_nav import log
from mkdocs_awesome_nav.nav.config import NavConfig
from mkdocs_awesome_nav.nav.context import Directory, MkdocsFilesContext, Page
from mkdocs_awesome_nav.nav.directory import NavDirectory
from mkdocs_awesome_nav.nav.page import NavPage
from mkdocs_awesome_nav.nav.section import NavSection
from mkdocs_awesome_nav.utils import create_absolute_pattern


class NavPattern:
    def __init__(self, pattern: str, *, path: PurePosixPath, config: NavConfig, ignore_no_matches: bool = False):
        self.pattern = pattern
        self.path = path
        self.config = config
        self._ignore_no_matches = ignore_no_matches

    def resolve(self, context: MkdocsFilesContext) -> list[NavPage | NavSection]:
        from mkdocs_awesome_nav.nav.resolve import resolve_in_priority_order

        matches = self._find_matches(context)

        if len(matches) == 0 and not self._ignore_no_matches:
            log.write(
                log.levels.no_matches or "warning",
                f"The nav item '{self.pattern}' doesn't match any files or directories",
                self.config.config_path,
            )

        items: list[NavPage | NavSection] = resolve_in_priority_order(matches, context)

        self._sort_items(items)

        return items

    def _find_matches(self, context: MkdocsFilesContext):
        absolut_pattern = create_absolute_pattern(self.path, self.pattern)
        matches: list[NavPage | NavDirectory] = []

        for file_object in context.unvisited:
            path = file_object.path.as_posix()
            if isinstance(file_object, Directory):
                path += "/"

            if glob.globmatch(
                path,
                absolut_pattern,
                flags=glob.GLOBSTAR | glob.EXTGLOB,
                exclude=self.config.ignore,
            ):
                if isinstance(file_object, Page):
                    matches.append(NavPage(file_object))
                elif isinstance(file_object, Directory):
                    directory_item = NavDirectory(file_object, parent_config=self.config)
                    if not directory_item.config.hide:
                        matches.append(directory_item)

        return matches

    def _sort_items(self, items: list[NavPage | NavSection]) -> None:
        key = self._sort_key_by_path
        if self.config.sort.by == "filename":
            key = self._sort_key_by_filename
        if self.config.sort.by == "title":
            key = self._sort_key_by_title

        if self.config.sort.type == "natural":
            natsort_flags = ns.GROUPLETTERS | ns.INT
            if self.config.sort.by != "title":
                natsort_flags |= ns.PATH
            if self.config.sort.ignore_case:
                natsort_flags |= ns.IGNORECASE
            key = cast(Any, natsort_keygen(key=key, alg=natsort_flags))
        elif self.config.sort.ignore_case:
            key = self._casefold_sort_key(key)

        items.sort(key=key, reverse=self.config.sort.direction == "desc")

        if self.config.sort.sections != "mixed":

            def _sort_key_type(item: NavPage | NavSection):
                if isinstance(item, NavSection):
                    return 2
                return 1

            items.sort(
                key=_sort_key_type,
                reverse=self.config.sort.sections == "first",
            )

    @staticmethod
    def _sort_key_by_path(item: NavPage | NavSection):
        return NavPattern._get_item_path(item).as_posix()

    @staticmethod
    def _sort_key_by_filename(item: NavPage | NavSection):
        path = NavPattern._get_item_path(item)
        return path.name, path.as_posix()

    @staticmethod
    def _sort_key_by_title(item: NavPage | NavSection):
        path = NavPattern._get_item_path(item)
        if isinstance(item, NavSection):
            return item.title, path.name, path.as_posix()

        markdown, metadata = get_data(item.page.file.content_string)
        return metadata.get("title") or item.page.path.name, path.name, path.as_posix()

    @staticmethod
    def _get_item_path(item: NavPage | NavSection):
        if isinstance(item, NavSection):
            return item.path
        return item.page.path

    @staticmethod
    def _casefold_sort_key(key: Callable[[NavPage | NavSection], tuple[str, str] | str]):
        def _casefolded_key(item: NavPage | NavSection):
            value = key(item)
            if isinstance(value, str):
                return value.casefold()
            return tuple([item.casefold() for item in value])

        return _casefolded_key



from mkdocs.config import Config, config_options
from mkdocs.config.defaults import MkDocsConfig
from mkdocs.plugins import BasePlugin, event_priority
from mkdocs.structure.files import Files
from mkdocs.structure.nav import Navigation, get_navigation
from mkdocs.structure.pages import Page

from mkdocs_awesome_nav import log
from mkdocs_awesome_nav.nav.context import MkdocsFilesContext
from mkdocs_awesome_nav.nav.directory import RootNavDirectory


class AwesomeNavConfig(Config):
    filename = config_options.Type(str, default=".nav.yml")
    logs = config_options.SubConfig(log.LogLevelsConfig)


class AwesomeNavPlugin(BasePlugin[AwesomeNavConfig]):
    @event_priority(100)
    def on_nav(self, nav: Navigation, config: MkDocsConfig, files: Files) -> Navigation:
        log.levels = self.config.logs

        if config.nav is not None:
            log.write(
                log.levels.nav_override or ("info" if "monorepo" in config.plugins else "warning"),
                "'nav' config from mkdocs.yml is being replaced with one generated by awesome-nav",
            )

        context = MkdocsFilesContext(files, self.config.filename)

        # clear pages created by mkdocs
        for file in files.documentation_pages():
            # check type for compatibility with material blog (preserve View objects)
            if type(file.page) is Page:
                file.page = None

        original_nav_config = config.nav
        config.nav = RootNavDirectory(context.root).resolve(context).to_mkdocs_config()
        nav = get_navigation(files, config)
        config.nav = original_nav_config

        return nav



from pathlib import PurePosixPath
from typing import TypeVar

T = TypeVar("T")


def flatten_list(items: list[list[T] | T]) -> list[T]:
    result: list[T] = []
    for item in items:
        if isinstance(item, list):
            result.extend(item)
        else:
            result.append(item)
    return result


def create_absolute_pattern(path: PurePosixPath, pattern: str) -> str:
    return (path / pattern).as_posix() + ("/" if pattern.endswith("/") else "")



from __future__ import annotations

from pathlib import PurePosixPath

from mkdocs_awesome_nav.nav.link import NavLink
from mkdocs_awesome_nav.nav.page import NavPage


class NavSection:
    def __init__(self, children: list[NavPage | NavLink | NavSection], *, path: PurePosixPath, title: str):
        self.children = children
        self.path = path
        self.title = title

    def to_mkdocs_config(self):
        return {self.title: [child.to_mkdocs_config() for child in self.children]}


class RootSection(NavSection):
    def to_mkdocs_config(self):
        return super().to_mkdocs_config()[self.title]



from typing import Optional, Type

from mkdocs.utils import dirname_to_title
from mkdocs.utils.meta import get_data
from pydantic import ValidationError
from yaml import YAMLError

from mkdocs_awesome_nav import log
from mkdocs_awesome_nav.log import format_log_message
from mkdocs_awesome_nav.nav.config import ConfigModel, NavConfig
from mkdocs_awesome_nav.nav.context import Directory, MkdocsFilesContext, Page
from mkdocs_awesome_nav.nav.link import NavLink
from mkdocs_awesome_nav.nav.page import NavPage
from mkdocs_awesome_nav.nav.section import NavSection, RootSection


class NavDirectory:
    def __init__(self, directory: Directory, *, parent_config: Optional[NavConfig] = None, title: Optional[str] = None):
        self._directory = directory
        self.path = directory.path
        self.config = self._load_config(parent_config)
        self._title = title or self.config.title

    def resolve(self, context: MkdocsFilesContext) -> NavSection | NavPage | list[NavSection | NavPage]:
        context.visit(self._directory)

        if self._title is None:
            self._title = self._generate_title(context)

        section = self._create_section(context)
        return self._flatten_section(section)

    def _load_config(self, parent_config: Optional[NavConfig]):
        config_file = self._directory.config_file
        if config_file is not None:
            try:
                return NavConfig.from_file(config_file, parent=parent_config)
            except YAMLError as e:
                log.write(
                    "error",
                    format_log_message("Parsing error", config_file.src_uri) + "\n" + str(e),
                )
            except ValidationError as e:
                lines = str(e).splitlines()
                lines[0] = format_log_message("Validation error", config_file.src_uri)
                log.write("error", "\n".join(lines))

        return NavConfig(ConfigModel(), directory_path=self.path, parent=parent_config)

    def _generate_title(self, context: MkdocsFilesContext) -> str:
        if self.config.preserve_directory_names:
            return self.path.name

        # Check if we should use index.md title
        if self.config.use_index_title:
            index_title = self._get_index_md_title(context)
            if index_title:
                return index_title

        return dirname_to_title(self.path.name)

    def _get_index_md_title(self, context: MkdocsFilesContext) -> Optional[str]:
        index_path = self.path / "index.md"
        file_object = context.get_by_path(index_path)
        if file_object and isinstance(file_object, Page):
            markdown, metadata = get_data(file_object.file.content_string)
            title = metadata.get("title")
            if title:
                return title
        return None

    def _flatten_section(self, section: NavSection) -> NavSection | NavPage | list[NavSection | NavPage]:
        if len(section.children) == 0:
            return []

        if self.config.flatten_single_child_sections and len(section.children) == 1:
            child = section.children[0]
            if isinstance(child, NavPage) or isinstance(child, NavSection):
                return child

        return section

    def _create_section(self, context: MkdocsFilesContext, *, section_type: Type[NavSection] = NavSection):
        from mkdocs_awesome_nav.nav.resolve import resolve_in_priority_order

        parsed_children = [item.parse(path=self.path, config=self.config, context=context) for item in self.config.nav]

        resolved_children: list[NavPage | NavSection | NavLink] = resolve_in_priority_order(parsed_children, context)

        return section_type(resolved_children, path=self.path, title=self._title)


class RootNavDirectory(NavDirectory):
    def __init__(self, directory: Directory):
        super().__init__(directory, title="root")

        if self.config.title is not None:
            log.write(
                log.levels.root_title or "warning",
                "'title' option has no effect at the top level",
                self.config.config_path,
            )
        if self.config.hide:
            log.write(
                log.levels.root_hide or "warning",
                "'hide' option has no effect at the top level",
                self.config.config_path,
            )

    def resolve(self, context: MkdocsFilesContext) -> NavSection:
        return super()._create_section(context, section_type=RootSection)



from __future__ import annotations

from functools import cached_property
from pathlib import PurePosixPath
from typing import Annotated, Literal, Optional

import yaml
from mkdocs.structure.files import File
from pydantic import BaseModel, Field, RootModel

from mkdocs_awesome_nav.nav.context import Directory, MkdocsFilesContext, Page
from mkdocs_awesome_nav.nav.link import NavLink
from mkdocs_awesome_nav.nav.page import NavPage
from mkdocs_awesome_nav.nav.section import NavSection
from mkdocs_awesome_nav.utils import create_absolute_pattern

NonEmptyString = Annotated[str, Field(min_length=1)]

inherit_token = "$inherit"


class NavConfig:
    def __init__(
        self,
        model: ConfigModel,
        *,
        directory_path: PurePosixPath,
        config_path: Optional[PurePosixPath] = None,
        parent: Optional[NavConfig] = None,
    ):
        self._model = model
        self._directory_path = directory_path
        self.config_path = config_path
        self._parent = parent

    @cached_property
    def nav(self) -> list[NavConfigItem]:
        nav = self._model.nav.copy()
        if self.append_unmatched:
            nav.append(NavConfigItemPatternOptions(glob="*", ignore_no_matches=True))
        return nav

    @cached_property
    def title(self) -> Optional[str]:
        return self._model.title

    @cached_property
    def hide(self) -> bool:
        return self._model.hide

    @cached_property
    def preserve_directory_names(self) -> bool:
        return (
            self._model.preserve_directory_names
            if self._model.preserve_directory_names is not None
            else self._parent.preserve_directory_names
            if self._parent is not None
            else False
        )

    @cached_property
    def flatten_single_child_sections(self) -> bool:
        return (
            self._model.flatten_single_child_sections
            if self._model.flatten_single_child_sections is not None
            else self._parent.flatten_single_child_sections
            if self._parent is not None
            else False
        )

    @cached_property
    def use_index_title(self) -> bool:
        return (
            self._model.use_index_title
            if self._model.use_index_title is not None
            else self._parent.use_index_title
            if self._parent is not None
            else False
        )

    @cached_property
    def sort(self) -> SortConfig:
        by = self._model.sort.by
        direction = self._model.sort.direction
        type = self._model.sort.type
        sections = self._model.sort.sections
        ignore_case = self._model.sort.ignore_case

        if self._parent:
            by = by or self._parent.sort.by
            direction = direction or self._parent.sort.direction
            type = type or self._parent.sort.type
            sections = sections or self._parent.sort.sections
            ignore_case = ignore_case or self._parent.sort.ignore_case

        return SortConfig(
            by=by or "path",
            direction=direction or "asc",
            type=type or "natural",
            sections=sections or "last",
            ignore_case=ignore_case or False,
        )

    @cached_property
    def ignore(self) -> list[str]:
        patterns = list()
        for pattern in self._model.ignore.patterns:
            if pattern == inherit_token:
                if self._parent is not None:
                    patterns.extend(self._parent.ignore)
            else:
                base_path = self._directory_path
                if pattern.startswith("/"):
                    pattern = pattern.removeprefix("/")
                else:
                    base_path /= "**"

                patterns.append(create_absolute_pattern(base_path, pattern))

        return patterns

    @cached_property
    def append_unmatched(self) -> bool:
        return (
            self._model.append_unmatched
            if self._model.append_unmatched is not None
            else self._parent.append_unmatched
            if self._parent is not None
            else False
        )

    @staticmethod
    def from_file(file: File, parent: Optional[NavConfig] = None) -> NavConfig:
        path = PurePosixPath(file.src_uri)
        return NavConfig(
            ConfigModel.from_yaml(file.content_string),
            directory_path=path.parent,
            parent=parent,
            config_path=path,
        )


class SortConfig(BaseModel):
    by: Optional[Literal["path", "filename", "title"]] = None
    direction: Optional[Literal["asc", "desc"]] = None
    type: Optional[Literal["natural", "alphabetical"]] = None
    sections: Optional[Literal["first", "last", "mixed"]] = None
    ignore_case: Optional[bool] = None


class IgnoreConfig(RootModel):
    root: Optional[NonEmptyString | list[NonEmptyString]] = None

    @property
    def patterns(self) -> list[str]:
        if self.root is None:
            return [inherit_token]
        if isinstance(self.root, list):
            return self.root
        return [self.root]


class NavConfigItemStr(RootModel):
    root: NonEmptyString

    def parse(self, *, path: PurePosixPath, config: NavConfig, context: MkdocsFilesContext):
        from mkdocs_awesome_nav.nav.directory import NavDirectory
        from mkdocs_awesome_nav.nav.pattern import NavPattern

        file_object = context.get_by_path(path / self.root)

        if isinstance(file_object, Page):
            return NavPage(file_object)
        elif isinstance(file_object, Directory):
            return NavDirectory(file_object, parent_config=config)
        else:
            return NavPattern(self.root, path=path, config=config)


class NavConfigItemDict(RootModel):
    root: Annotated[dict[NonEmptyString, NonEmptyString | list[NavConfigItem]], Field(min_length=1, max_length=1)]

    def parse(self, *, path: PurePosixPath, context: MkdocsFilesContext, config: NavConfig):
        from mkdocs_awesome_nav.nav.directory import NavDirectory

        (key, value) = list(self.root.items())[0]
        if isinstance(value, str):
            file_object = context.get_by_path(path / value)
            if isinstance(file_object, Page):
                return NavPage(file_object, title=key)
            elif isinstance(file_object, Directory):
                return NavDirectory(file_object, parent_config=config, title=key)
            else:
                return NavLink(value, title=key)
        elif isinstance(value, list):
            children = [item.parse(path=path, config=config, context=context) for item in value]
            return NavSection(children, path=path, title=key)


class NavConfigItemPatternOptions(BaseModel):
    glob: NonEmptyString
    flatten_single_child_sections: Optional[bool] = None
    preserve_directory_names: Optional[bool] = None
    sort: SortConfig = SortConfig()
    ignore: IgnoreConfig = IgnoreConfig()
    append_unmatched: Optional[bool] = None
    ignore_no_matches: bool = False

    def parse(self, *, path: PurePosixPath, context: MkdocsFilesContext, config: NavConfig):
        from mkdocs_awesome_nav.nav.pattern import NavPattern

        local_config = NavConfig(
            ConfigModel(
                sort=self.sort,
                flatten_single_child_sections=self.flatten_single_child_sections,
                preserve_directory_names=self.preserve_directory_names,
                ignore=self.ignore,
                append_unmatched=self.append_unmatched,
            ),
            directory_path=path,
            config_path=config.config_path,
            parent=config,
        )

        return NavPattern(self.glob, path=path, config=local_config, ignore_no_matches=self.ignore_no_matches)


NavConfigItem = NavConfigItemStr | NavConfigItemDict | NavConfigItemPatternOptions


class ConfigModel(BaseModel):
    title: Optional[NonEmptyString] = None
    hide: bool = False
    flatten_single_child_sections: Optional[bool] = None
    preserve_directory_names: Optional[bool] = None
    use_index_title: Optional[bool] = None
    sort: SortConfig = SortConfig()
    ignore: IgnoreConfig = IgnoreConfig()
    nav: list[NavConfigItem] = [
        NavConfigItemPatternOptions(glob="@(index.md|README.md)", ignore_no_matches=True),
        NavConfigItemPatternOptions(glob="*", ignore_no_matches=True),
    ]
    append_unmatched: Optional[bool] = None

    @staticmethod
    def from_yaml(contents: str) -> ConfigModel:
        data = yaml.safe_load(contents) or {}
        return ConfigModel(**data)



from operator import attrgetter
from typing import TypeVar, cast

from mkdocs_awesome_nav.nav.context import MkdocsFilesContext
from mkdocs_awesome_nav.nav.directory import NavDirectory
from mkdocs_awesome_nav.nav.link import NavLink
from mkdocs_awesome_nav.nav.page import NavPage
from mkdocs_awesome_nav.nav.pattern import NavPattern
from mkdocs_awesome_nav.nav.section import NavSection
from mkdocs_awesome_nav.utils import flatten_list

TInput = TypeVar("TInput", bound=NavPage | NavPattern | NavDirectory | NavLink | NavSection)
TResult = TypeVar("TResult", bound=NavPage | NavSection | NavLink)


def resolve_in_priority_order(items: list[TInput], context: MkdocsFilesContext) -> list[TResult]:
    queue = _create_resolve_queue(items)
    for to_resolve in sorted(queue, key=attrgetter("priority")):
        to_resolve.resolve(context)

    return _flatten_result(cast(list[TResult | list[TResult]], items))


def _create_resolve_queue(items: list[TInput]):
    for index, item in enumerate(items):
        if isinstance(item, NavSection):
            yield from _create_resolve_queue(item.children)
        elif isinstance(item, NavPage) or isinstance(item, NavDirectory) or isinstance(item, NavPattern):
            yield _ItemToBeResolved(item, result=items, result_index=index)


def _flatten_result(items: list[TResult | list[TResult]]):
    result = flatten_list(items)
    for item in result:
        if isinstance(item, NavSection):
            item.children = _flatten_result(cast(list[TResult | list[TResult]], item.children))
    return result


class _ItemToBeResolved:
    def __init__(
        self,
        item: NavPage | NavDirectory | NavPattern,
        result: list,
        result_index: int,
    ):
        self._item = item
        self._result = result
        self._result_index = result_index

    @property
    def priority(self):
        if isinstance(self._item, NavPage):
            return 1, self._result_index
        if isinstance(self._item, NavDirectory):
            return 2, -len(self._item.path.parts), self._result_index
        if isinstance(self._item, NavPattern):
            return 3, self._result_index

    def resolve(self, context: MkdocsFilesContext):
        self._result[self._result_index] = self._item.resolve(context)



from __future__ import annotations

from typing import Optional

from mkdocs_awesome_nav.nav.context import MkdocsFilesContext, Page


class NavPage:
    def __init__(self, page: Page, *, title: Optional[str] = None):
        self.page = page
        self.title = title

    def resolve(self, context: MkdocsFilesContext) -> NavPage:
        context.visit(self.page)
        return self

    def to_mkdocs_config(self):
        path = self.page.path.as_posix()

        if self.title:
            return {self.title: path}

        return path



from pathlib import PurePosixPath
from typing import Optional

from mkdocs.structure.files import File, Files, InclusionLevel


class Page:
    def __init__(self, file: File):
        self.path = PurePosixPath(file.src_uri)
        self.file = file


class Directory:
    def __init__(self, path: PurePosixPath, config_file: Optional[File]):
        self.path = path
        self.config_file = config_file


FileObject = Page | Directory


class MkdocsFilesContext:
    def __init__(self, mkdocs_files: Files, config_filename: str):
        self._mkdocs_files = mkdocs_files
        self._config_filename = config_filename

        self.root = self._create_directory(PurePosixPath("."))

        self._by_path: dict[PurePosixPath, FileObject] = dict()
        self.unvisited: set[FileObject] = set()

        for file in self._mkdocs_files.documentation_pages(inclusion=InclusionLevel.is_in_nav):
            page = Page(file)
            self._add_file_object(page)

            for parent_path in page.path.parents:
                if parent_path not in self._by_path:
                    directory = self._create_directory(parent_path)
                    self._add_file_object(directory)

    def get_by_path(self, path: PurePosixPath) -> Optional[FileObject]:
        # we go through get_file_from_path because the i18n plugin overrides it to resolve file.md into file.en.md
        mkdocs_file = self._mkdocs_files.get_file_from_path(path.as_posix())
        if mkdocs_file is not None:
            return self._by_path.get(PurePosixPath(mkdocs_file.src_uri), None)
        return self._by_path.get(path, None)

    def visit(self, item: FileObject) -> None:
        if item in self.unvisited:
            self.unvisited.remove(item)

    def _add_file_object(self, file_object: FileObject):
        self._by_path[file_object.path] = file_object
        self.unvisited.add(file_object)

    def _create_directory(self, path: PurePosixPath):
        config_file_path = path / self._config_filename
        config_file = self._mkdocs_files.get_file_from_path(config_file_path.as_posix())
        return Directory(path, config_file)



from pathlib import PurePosixPath
from typing import Optional, Literal

from mkdocs.config import config_options, Config
from mkdocs.plugins import get_plugin_logger

_logger = get_plugin_logger("awesome-nav")


LogLevel = Literal["info", "warning", "error"]


def _log_level_choice():
    return config_options.Optional(config_options.Choice[LogLevel](["info", "warning", "error"]))


class LogLevelsConfig(Config):
    nav_override = _log_level_choice()
    root_title = _log_level_choice()
    root_hide = _log_level_choice()
    no_matches = _log_level_choice()


levels = LogLevelsConfig()


def format_log_message(message: str, path: Optional[PurePosixPath | str] = None) -> str:
    if path is None:
        return message
    return f"{message} [{path}]"


def write(level: LogLevel, message: str, path: Optional[PurePosixPath | str] = None):
    formatted = format_log_message(message, path)
    match level:
        case "info":
            _logger.info(formatted)
        case "warning":
            _logger.warning(formatted)
        case "error":
            _logger.error(formatted)
